#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import os
from typing import Optional

import yaml

from gurglefish import DriverManager
from gurglefish.context import Context
from gurglefish.objects.connections import Connections, ConnectionConfig
from gurglefish.sfapi import SFClient
import logging.config

_log = logging.getLogger('main')


def setup_env(envname) -> Optional[Context]:

    logconfig = load_log_config()
    logging.config.dictConfig(logconfig)

    mde = Connections()
    env: ConnectionConfig = mde.get_db_env(envname)
    if env is None:
        _log.error(f'Configuration for {envname} not found')
        exit(1)

    sf = SFClient()
    try:
        sf.login(env.consumer_key, env.consumer_secret, env.login, env.password, env.authurl)
    except Exception as ex:
        _log.error(f'Unable to connect to {env.authurl} as {env.login}: {str(ex)}')
        return None

    return Context(envname, env, get_db_connection(envname), sf)


def get_db_connection(envname: str) -> DriverManager.DbDriverMeta:
    mde = Connections()
    env: ConnectionConfig = mde.get_db_env(envname)
    if env is None:
        _log.error(f'Configuration for {envname} not found')
        exit(1)

    driver = DriverManager.Manager().get_driver(env.dbvendor)
    driver.connect(env)
    return driver


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, datetime.datetime):
        serial = obj.isoformat()
        return serial
    return str(obj)


def dict_list_to_dict(alist, keyfield):
    assert(keyfield is not None)
    assert(alist is not None)

    result = dict()
    for item in alist:
        key = item[keyfield]
        result[key] = item
    return result


def sf_timestamp(t: datetime):
    s = t.isoformat()[0:19]
    s += '+00:00'
    return s


def parse_timestamp(t):
    return datetime.datetime.strptime(t[0:19], '%Y-%m-%dT%H:%M:%S')


def load_file_items(filename):
    with open(filename, 'r') as f:
        line_list = f.readlines()
        stripped_list = [line.strip() for line in line_list if len(line) > 0]
        return stripped_list


def make_arg_list(args_list):
    processed_args = []
    for arg in args_list:
        if len(arg) == 0:
            continue
        if arg.startswith('@'):
            processed_args.extend(load_file_items(arg[1:]))
        else:
            processed_args.append(arg)
    return processed_args


def load_log_config():
    path = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(path, 'logging.yml'), 'r') as configfile:
        _logconfig = yaml.load(configfile.read(), Loader=yaml.FullLoader)
        return _logconfig



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.
import logging
import json
import operator
import time
from typing import Dict, Optional

import requests
from fastcache import lru_cache

from gurglefish.objects.sobject import SObjectFields

MAX_BATCH_SIZE = 100
_API_VERSION = '44.0'


class SFQueryTooLarge(Exception):
    def __init__(self):
        super().__init__(self)


class JobBatch:

    def __init__(self, batchinfo: Dict, parent):
        self.parent = parent
        self.batch_id = batchinfo['id']
        self.batchinfo = batchinfo

    @property
    def state(self) -> str:
        return self.batchinfo['state']

    @property
    def id(self) -> str:
        return self.batch_id

    def get_results(self):
        response = self.parent.client.get(
            f'{self.parent.service_url}/services/async/{_API_VERSION}/job/{self.parent.job_id}/batch/{self.batch_id}/result')
        response.raise_for_status()
        self.client.calls += 1

        result = json.loads(response.text)
        for resultid in result:
            response = self.parent.client.get(
                f'{self.parent.service_url}/services/async/{_API_VERSION}/job/{self.parent.job_id}/batch/{self.batch_id}/result/{resultid}',
                stream=True)
            self.client.calls += 1
            doc = ''
            for chunk in response.iter_lines():
                if chunk is None:
                    continue
                chunk = chunk.decode('utf-8')
                if chunk[0] == '[':
                    doc = '{'
                elif chunk[0] == '}':
                    doc += '}'
                    yield json.loads(doc)
                    doc = '{'
                else:
                    doc += chunk

    def refresh(self):
        if self.state != 'Completed':
            response = self.parent.client.get(
                f'{self.parent.service_url}/services/async/{_API_VERSION}/job/{self.parent.job_id}/batch/{self.batch_id}')
            self.client.calls += 1
            response.raise_for_status()
            self.batchinfo = json.loads(response.text)
            return True
        return False

    def wait_for_start(self, timeout: int = 180) -> bool:
        for i in range(0, timeout, 30):
            time.sleep(30)
            self.refresh()
            if self.state in ('Completed', 'NotProcessed'):
                return True
            if self.state == 'Failed':
                return False
        return False


class BulkJob:
    JOB_OP_QUERY = 'query'
    JOB_OP_INSERT = 'insert'
    JOB_OP_UPDATE = 'update'
    JOB_TYPE_CSV = 'CSV'
    JOB_TYPE_XML = 'XML'
    JOB_TYPE_JSON = 'JSON'
    CONCUR_PARALLEL = 'Parallel'
    CONCUR_SERIAL = 'Serial'

    def __init__(self, jobinfo: Dict, client, service_url):
        self.job_id = jobinfo['id']
        self.jobinfo = jobinfo
        self.client = client
        self.service_url = service_url
        self.pending: [JobBatch] = list()
        self.complete: [str] = list()

    def create_batch(self, payload):
        response = self.client.post(f'{self.service_url}/services/async/{_API_VERSION}/job/{self.job_id}/batch',
                                    data=json.dumps(payload, indent=4),
                                    headers={'Content-Type': 'application/json; charset=UTF-8'})
        self.client.calls += 1
        response.raise_for_status()
        result = json.loads(response.text)
        return JobBatch(result, self)

    @property
    def batch_count(self) -> int:
        return self.jobinfo['numberBatchesTotal'] + self.jobinfo['numberBatchesQueued']

    def bulk_query(self, soql: str):
        response = self.client.post(f'{self.service_url}/services/async/{_API_VERSION}/job/{self.job_id}/batch',
                                    data=soql + ' ',
                                    headers={'Content-Type': 'application/json; charset=UTF-8'})
        self.client.calls += 1
        response.raise_for_status()
        result = json.loads(response.text)
        batch = JobBatch(result, self)
        self.pending.append(batch)
        return batch

    def is_done(self) -> bool:
        self.refresh()
        self.get_batches()
        all = self.jobinfo['numberBatchesQueued'] + self.jobinfo['numberBatchesInProgress'] + \
              self.jobinfo['numberBatchesFailed'] + self.jobinfo['numberBatchesCompleted']

        return len(self.complete) >= all and len(self.pending) == 0 and self.jobinfo['numberBatchesQueued'] == 0

#        return 0 < all < self.jobinfo['numberBatchesTotal'] \
#               and len(self.complete) < len(self.pending) \
#               and self.jobinfo['numberBatchesQueued'] == 0

    #
    # Returns:
    # {
    #   "apexProcessingTime": 0,
    #   "apiActiveProcessingTime": 0,
    #   "apiVersion": 36.0,
    #   "concurrencyMode": "Parallel",
    #   "contentType": "JSON",
    #   "createdById": "005D0000001b0fFIAQ",
    #   "createdDate": "2015-12-15T20:45:25.000+0000",
    #   "id": "750D00000004SkGIAU",
    #   "numberBatchesCompleted": 0,
    #   "numberBatchesFailed": 0,
    #   "numberBatchesInProgress": 0,
    #   "numberBatchesQueued": 0,
    #   "numberBatchesTotal": 0,
    #   "numberRecordsFailed": 0,
    #   "numberRecordsProcessed": 0,
    #   "numberRetries": 0,
    #   "object": "Account",
    #   "operation": "insert",
    #   "state": "Open",
    #   "systemModstamp": "2015-12-15T20:45:25.000+0000",
    #   "totalProcessingTime": 0
    # }
    #
    def refresh(self):
        response = self.client.get(f'{self.service_url}/services/async/{_API_VERSION}/job/{self.job_id}')
        self.client.calls += 1
        response.raise_for_status()
        self.jobinfo = json.loads(response.text)

    #
    # Returns:
    #
    # {
    #   "batchInfo": [
    #   {
    #     "apexProcessingTime": 0,
    #     "apiActiveProcessingTime": 0,
    #     "createdDate": "2015-12-15T21:56:43.000+0000",
    #     "id": "751D00000004YGZIA2",
    #     "jobId": "750D00000004SkVIAU",
    #     "numberRecordsFailed": 0,
    #     "numberRecordsProcessed": 0,
    #     "state": "Queued",
    #     "systemModstamp": "2015-12-15T21:57:19.000+0000",
    #     "totalProcessingTime": 0
    #   }
    #   ]
    # }
    #
    def get_batches(self) -> [Dict]:
        response = self.client.get(f'{self.service_url}/services/async/{_API_VERSION}/job/{self.job_id}/batch')
        self.client.calls += 1
        response.raise_for_status()
        result = json.loads(response.text)
        self.pending: [JobBatch] = list()
        for bi in result['batchInfo']:
            if bi['id'] not in self.complete:
                if bi['state'] in ('Queued', 'Completed', 'InProgress'):
                    self.pending.append(JobBatch(bi, self))
                else:
                    self.release_batch(bi['id'])
                    print('**** unhandled state ' + bi['state'])

    def release_batch(self, batch_id: str):
        self.complete.append(batch_id)

    def get_completed_batch(self) -> Optional[JobBatch]:
        self.get_batches()
        for batch in self.pending:
            if batch.state == 'Completed':
                return batch
        return None

    def close(self):
        response = self.client.post(f'{self.service_url}/services/async/{_API_VERSION}/job/{self.job_id}',
                                    data='{"state":"Closed"}')
        self.client.calls += 1
        response.raise_for_status()
        result = json.loads(response.text)
        return result


class SFClient:

    def __init__(self):
        self.log = logging.getLogger('salesforce')
        self.access_token = None
        self.service_url = None
        self.client = None
        self._username = None
        self.calls = 0

    def login(self, consumer_key, consumer_secret, username, password, server_url):
        self._username = username
        payload = {'grant_type': 'password',
                   'username': username,
                   'password': password,
                   'client_id': consumer_key,
                   'client_secret': consumer_secret
                   }
        rsp = requests.post(server_url + '/services/oauth2/token', data=payload,
                            headers={'content-type': 'application/x-www-form-urlencoded'})
        payload = json.loads(rsp.text)
        if 'error' in payload:
            raise Exception(payload['error_description'])
        # self.log.debug('payload=%s' % (rsp.text,))
        self.construct(payload['access_token'], payload['instance_url'])

    def construct(self, token, server_url):
        self.access_token = token
        self.service_url = server_url
        self.client = requests.Session()
        self.client.headers.update({'Authorization': 'OAuth ' + token,
                                    'X-SFDC-Session': token,
                                    'Content-Type': 'application/json; charset=UTF-8',
                                    'Accept-Encoding': 'gzip, compress, deflate', 'Accept-Charset': 'utf-8'})

    def close(self):
        pass

    @lru_cache(maxsize=3, typed=False)
    def get_sobject_definition(self, name: str) -> Dict:
        sobject_doc = self._get('sobjects/{}/describe'.format(name), {})
        return sobject_doc

    @lru_cache(maxsize=1, typed=False)
    def get_sobject_list(self) -> [Dict]:
        payload = self._get('sobjects/', {})
        return payload['sobjects']

    @lru_cache(maxsize=3, typed=False)
    def get_field_list(self, sobject_name: str) -> SObjectFields:
        fielddef = self._get('sobjects/%s/describe/' % (sobject_name,), {})
        fieldlist = fielddef['fields']
        fieldlist.sort(key=operator.itemgetter('name'))
        return SObjectFields(fieldlist)

    def dump_ids(self, sobject_name, output_filename: str):
        with open(output_filename, 'w') as out:
            for rec in self.query(f'select Id from {sobject_name} order by Id'):
                out.write(rec['Id'][0:15] + '\n')

    def record_count(self, sobject: str, query_filter: str = None):
        soql = 'select count() from ' + sobject
        if query_filter:
            soql += ' where ' + query_filter
        fullurl = f'{self.service_url}/services/data/v{_API_VERSION}/query/'
        self.log.debug('get %s', fullurl)
        response = self.client.get(fullurl, params={'q': soql})
        result_payload = response.json()
        if response.status_code != 200:
            self.log.error(f'query error {response.status_code}, {response.reason}')
            self.log.error(result_payload)
            return
        self.calls += 1
        return result_payload['totalSize']

    def query(self, soql: str, include_deleted=False):
        resource = 'queryAll' if include_deleted else 'query'
        fullurl = f'{self.service_url}/services/data/v{_API_VERSION}/{resource}/'
        #
        # Need to make specific changes to soql to avoid upsetting Salesforce and
        # using requests built-in escaping causes problems.
        #
        soql = soql.replace('+', '%2b').replace('\n', '').replace('\r', '').replace(' ', '+')
        response = self.client.get(fullurl, params=f'q={soql}')
        result_payload = response.text
        if response.status_code == 431:
            raise SFQueryTooLarge()
        self.calls += 1
        if response.status_code != 200:
            self.log.error(f'query error {response.status_code}, {response.reason}')
            self.log.error(result_payload)
            return
        data = json.loads(result_payload)
        recs = data['records']
        for rec in recs:
            yield (rec)
        while 'nextRecordsUrl' in data:
            next_records_url = data['nextRecordsUrl']
            if next_records_url:
                response = self.client.get('%s%s' % (self.service_url, next_records_url))
                self.calls += 1
                txt = response.text
                if isinstance(txt, str):
                    result_payload = txt
                else:
                    result_payload = str(txt, 'utf-8')
                data = json.loads(result_payload)
                recs = data['records']
                for rec in recs:
                    yield (rec)
            else:
                break

    def _get(self, url, url_params):
        fullurl = f'{self.service_url}/services/data/v{_API_VERSION}/{url}'
        response = self.client.get(fullurl, params=url_params)
        result_payload = response.text
        self.calls += 1
        if response.status_code != 200:
            self.log.debug('get %s', fullurl)
        response.raise_for_status()
        data = json.loads(result_payload)
        return data

    def _post(self, url, url_params):
        fullurl = f'{self.service_url}/services/data/v{_API_VERSION}/{url}'
        response = self.client.post(fullurl, params=url_params)
        self.calls += 1
        if response.status_code != 200:
            self.log.debug('post %s', fullurl)
        result_payload = response.text
        response.raise_for_status()
        data = json.loads(result_payload)
        return data

    def add_header(self, name: str, val: str):
        self.client.headers[name] = val

    def drop_header(self, name: str):
        if name in self.client.headers:
            del self.client.headers[name]

    def create_job(self, op, sobject_name, content_type='JSON', concur_mode=BulkJob.CONCUR_PARALLEL):
        payload = {"operation": op, "object": sobject_name, "contentType": content_type,
                   "concurrencyMode": concur_mode}
        url = f'{self.service_url}/services/async/{_API_VERSION}/job'
        self.client.headers['Content-Type'] = 'application/json; charset=UTF-8'
        response = self.client.post(url, data=json.dumps(payload),
                                    headers={'Content-Type': 'application/json; charset=UTF-8'})
        self.calls += 1
        response.raise_for_status()
        result = response.json()
        if result['state'] != 'Open':
            raise Exception('Invalid job state: {}'.format(result['state']))
        return BulkJob(result, self.client, self.service_url)

    def bulk_query(self, sobject: str, soql: str, job_id=None, timeout=600):

        if job_id is None:
            job = self.create_job(BulkJob.JOB_OP_QUERY, sobject)
            main_batch = job.bulk_query(soql + ' ')
            job.close()
            self.log.info(f'Waiting on bulk query job to start, timeout is {timeout} seconds')

            # suppress annoying requests debug logging
            logging.getLogger('chardet.charsetprober').setLevel(logging.INFO)

            if main_batch.wait_for_start(timeout):
                counter = 0
                batches = job.batch_count
                job.release_batch(main_batch.id)
                while not job.is_done():
                    completed = job.get_completed_batch()
                    if completed is not None:
                        counter += 1
                        self.log.debug(f'Processing batch {counter} of {batches}')
                        for result in completed.get_results():
                            del result['attributes']
                            yield result
                        job.release_batch(completed.id)
                    else:
                        time.sleep(30)
            else:
                self.log.error('Timed out waiting for bulk query job to start')
        else:
            job = BulkJob({'id': job_id, 'state': 'Closed'}, self.client, self.service_url)
            counter = 1
            while not job.is_done():
                total = job.batch_count
                completed = job.get_completed_batch()
                if completed is not None:
                    self.log.debug(f'Processing batch {completed.batch_id} - {counter} of {total}')
                    for result in completed.get_results():
                        del result['attributes']
                        yield result
                    job.release_batch(completed.id)
                    counter += 1
                else:
                    time.sleep(20)



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
import pkgutil
from abc import ABCMeta, abstractmethod
from typing import List, Optional, Dict

from gurglefish.objects.connections import ConnectionConfig
from gurglefish.objects.sobject import ColumnMap, SObjectFields


class GetDbTablesResult(object):

    def __init__(self, name):
        self.tablename = name

    @property
    def tablename(self):
        return self.name

    @tablename.setter
    def tablename(self, name):
        self.name = name


class DbNativeExporter(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def soql(self) -> str:
        pass

    @abstractmethod
    def write(self, record: Dict):
        pass

    @abstractmethod
    def close(self):
        pass


class DbDriverMeta(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def connect(self, env: ConnectionConfig):
        pass

    @abstractmethod
    def create_exporter(self, sobject_name: str, ctx, just_sample=False, timestamp=None) -> DbNativeExporter:
        pass

    @abstractmethod
    def get_db_tables(self)-> List[GetDbTablesResult]:
        pass

    @abstractmethod
    def table_exists(self, table_name: str):
        pass

    @abstractmethod
    def get_db_columns(self, table_name: str):
        pass

    @abstractmethod
    def dump_ids(self, table_name: str, output_filename: str):
        pass

    @abstractmethod
    def make_create_table(self, fields: SObjectFields, sobject_name: str):
        pass

    @abstractmethod
    def make_select_statement(self, field_names: [str], sobject_name: str) -> str:
        pass

    @abstractmethod
    def exec_ddl(self, ddl: str):
        pass

    @abstractmethod
    def max_timestamp(self, tablename: str):
        pass

    @abstractmethod
    def format_for_export(self, trec, tablefields, fieldmap):
        pass

    @abstractmethod
    def make_transformer(self, sobject_name, table_name: str, fieldlist):
        pass

    @abstractmethod
    def maintain_indexes(self, sobject_name: str, field_defs):
        pass

    @abstractmethod
    def record_count(self, table_name: str):
        pass

    @abstractmethod
    def get_table_fields(self, table_name: str):
        pass

    @abstractmethod
    def delete(self, cur, table_name: str, key: str):
        pass

    @abstractmethod
    def upsert(self, cur, table_name: str, trec: dict, journal=None):
        pass

    @abstractmethod
    def import_native(self, tablename: str):
        pass

    @abstractmethod
    def export_native(self, table_name, output_path):
        pass

    @abstractmethod
    def start_sync_job(self):
        pass

    @abstractmethod
    def finish_sync_job(self, jobid):
        pass

    @abstractmethod
    def insert_sync_stats(self, jobid, table_name, sync_start, sync_end, sync_since, inserts, updates, deletes, api_calls):
        pass

    @abstractmethod
    def clean_house(self, date_constraint: datetime):
        pass

    @abstractmethod
    def alter_table_drop_columns(self, drop_field_names: [str], sobject_name: str):
        pass

    @abstractmethod
    def alter_table_add_columns(self, new_field_defs, sobject_name: str) -> [ColumnMap]:
        pass

    @abstractmethod
    def cursor(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def commit(self):
        pass

    @abstractmethod
    def rollback(self):
        pass


class Manager(object):

    def __init__(self):
        self._res = {}
        path = os.path.dirname(os.path.abspath(__file__))
        modules = pkgutil.iter_modules(path=[os.path.join(path, 'drivers')])
        for finder, mod_name, ispkg in modules:
            toload = finder.find_module(mod_name)
            mod = toload.load_module(mod_name)
            cls = getattr(mod, 'Driver')
            self._res[mod_name] = cls

    def get_driver(self, driver_name) -> Optional[DbDriverMeta]:
        if driver_name in self._res:
            return self._res[driver_name]()
        return None



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.
from gurglefish import config
from gurglefish.DriverManager import DbDriverMeta
from gurglefish.FileManager import FileManager
from gurglefish.objects.connections import ConnectionConfig
from gurglefish.sfapi import SFClient


class Context:

    def __init__(self, envname: str, env: ConnectionConfig, dbdriver: DbDriverMeta, sfclient: SFClient):
        self.env = env
        self.envname = envname
        self.driver = dbdriver
        self.sfapi = sfclient
        self.filemgr = FileManager(config.storagedir, self.env.id)

    @property
    def config_env(self) -> ConnectionConfig:
        return self.env

    @property
    def dbdriver(self) -> DbDriverMeta:
        return self.driver

    @property
    def sfclient(self) -> SFClient:
        return self.sfapi



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.

import gzip
import json
import os
import sys
from typing import Optional

from gurglefish.objects.files import LocalTableConfig
from gurglefish.objects.sobject import ColumnMap

from gurglefish.sfapi import SObjectFields


class FileManager(object):

    def __init__(self, basedir, envname):
        self.basedir = basedir
        self.envname = envname
        self.schemadir = os.path.join(basedir, 'db', envname, 'schema')
        self.exportdir = os.path.join(basedir, 'db', envname, 'export')

        if not os.path.exists(basedir):
            print(f'Gurglefish root director {basedir} does not exists')
            sys.exit(1)

        try:
            os.makedirs(self.schemadir, exist_ok=True)
        except PermissionError as pex:
            print(f'Permission denied creating {self.schemadir}')
            sys.exit(1)

        try:
            os.makedirs(self.exportdir, exist_ok=True)
        except PermissionError as pex:
            print(f'Permission denied creating {self.exportdir}')
            sys.exit(1)

    def create_journal(self, sobject_name):
        f = gzip.open(os.path.join(self.exportdir, '{}_journal.log.gz'.format(sobject_name)), 'wb')
        return f

    def get_global_filters(self):
        try:
            with open(os.path.join(self.basedir, 'global-filters.txt'), 'r') as filterfile:
                return filterfile.readlines()
        except:
            pass
        return []

    def get_filters(self):
        try:
            with open(os.path.join(self.basedir, 'db', self.envname, 'filters.txt', 'r')) as filterfile:
                return filterfile.readlines()
        except:
            pass
        return []

    def get_schema_list(self):
        return os.listdir(self.schemadir)

    def get_export_list(self):
        return os.listdir(self.exportdir)

    def load_translate_handler(self, sobject_name):
        import importlib.machinery
        loader = importlib.machinery.SourceFileLoader(sobject_name, os.path.join(self.schemadir, sobject_name, '{}_Transform.py'.format(sobject_name)))
        handler = loader.load_module(sobject_name)
        return handler

    def get_sobject_fields(self, sobject_name: str) -> Optional[SObjectFields]:
        os.makedirs(os.path.join(self.schemadir, sobject_name), exist_ok=True)
        filename = os.path.join(self.schemadir, sobject_name, sobject_name + '.json')
        try:
            with open(filename, 'r') as jsonfile:
                return SObjectFields(json.load(jsonfile))
        except Exception:
            return None

    def save_sobject_fields(self, sobject_name: str, fields: SObjectFields):
        sobject_name = sobject_name.lower()
        os.makedirs(os.path.join(self.schemadir, sobject_name), exist_ok=True)
        with open(os.path.join(self.schemadir, sobject_name, '{}.json'.format(sobject_name)), 'w') as mapfile:
            mapfile.write(json.dumps(fields.values_exportable(), indent=4))

    def get_configured_tables(self) -> [LocalTableConfig]:
        try:
            with open(os.path.join(self.basedir, 'db', self.envname, 'config.json'), 'r') as configfile:
                return [LocalTableConfig(t) for t in json.load(configfile)['configuration']['sobjects']]
        except FileNotFoundError:
            return None

    def save_configured_tables(self, new_config: [LocalTableConfig]):
        config = {'configuration': {'sobjects': []}}
        try:
            with open(os.path.join(self.basedir, 'db', self.envname, 'config.json'), 'r') as configfile:
                config = json.load(configfile)
        except FileNotFoundError:
            pass
        config['configuration']['sobjects'] = [t.dict for t in new_config]
        with open(os.path.join(self.basedir, 'db', self.envname, 'config.json'), 'w') as configfile:
            configfile.write(json.dumps(config, indent=4))

    def get_sobject_map(self, sobject_name: str) -> [ColumnMap]:
        sobject_name = sobject_name.lower()
        with open(os.path.join(self.schemadir, sobject_name, '{}_map.json'.format(sobject_name)), 'r') as mapfile:
            return [ColumnMap(f) for f in json.load(mapfile)]

    def save_sobject_map(self, sobject_name: str, fieldmap: [ColumnMap]):
        sobject_name = sobject_name.lower()
        os.makedirs(os.path.join(self.schemadir, sobject_name), exist_ok=True)
        with open(os.path.join(self.schemadir, sobject_name, '{}_map.json'.format(sobject_name)), 'w') as mapfile:
            mapfile.write(json.dumps([f.as_dict() for f in fieldmap], indent=4))

    def get_sobject_query(self, sobject_name: str):
        sobject_name = sobject_name.lower()
        with open(os.path.join(self.schemadir, sobject_name, 'query.soql'), 'r') as queryfile:
            return queryfile.read()

    def save_table_create(self, sobject_name: str, sql: str):
        sobject_name = sobject_name.lower()
        os.makedirs(os.path.join(self.schemadir, sobject_name), exist_ok=True)
        with open(os.path.join(self.schemadir, sobject_name, '{}.sql'.format(sobject_name)), 'w') as schemafile:
            schemafile.write(sql)

    def save_sobject_transformer(self, sobject_name: str, xformr: str):
        sobject_name = sobject_name.lower()
        os.makedirs(os.path.join(self.schemadir, sobject_name), exist_ok=True)
        with open(os.path.join(self.schemadir, sobject_name, '{}_Transform.py'.format(sobject_name)),
                  'w') as parserfile:
            parserfile.write(xformr)

    def save_sobject_query(self, sobject_name: str, soql: str):
        sobject_name = sobject_name.lower()
        os.makedirs(os.path.join(self.schemadir, sobject_name), exist_ok=True)
        with open(os.path.join(self.schemadir, sobject_name, 'query.soql'), 'w') as queryfile:
            queryfile.write(soql)



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.
from datetime import datetime, time, date

import decimal


def id(rec, name, fieldlen):
    if name in rec and rec[name] != None:
        if len(rec[name]) > 15:
            return rec[name][0:15]
        return rec[name]
    return None


def inte(rec, name, fieldlen):
    if name in rec and rec[name] is not None:
        return rec[name]
    return None


def bl(rec, name, fieldlen):
    if name in rec and rec[name] is not None:
        return rec[name]
    return None


def dt(rec, name, fieldlen):
    if name in rec and rec[name] is not None:
        return py_date(rec[name])
    return None


def tm(rec, name, fieldlen):
    if name in rec and rec[name] is not None:
        return py_time(rec[name])
    return None


def ts(rec, name, fieldlen):
    if name in rec and rec[name] is not None:
        return py_timestamp(rec[name])
    return None


def db(rec, name, fieldlen):
    if name in rec and rec[name] is not None:
        d = decimal.Decimal(rec[name])
        s = str(d)
        if 0 < fieldlen < len(s):
            # truncate
            return float(s[0:fieldlen])
        return rec[name]
    return None


def st(rec, name, fieldlen=0):
    if name in rec and rec[name] is not None:
        node = rec[name]
        return scrub(node[0:fieldlen])
    return None


def stsub(rec, name, subname, fieldlen=0):
    if name in rec and rec[name] is not None:
        node = rec[name]
        if subname in node:
            val = node[subname][0:fieldlen]
            return scrub(val)
        return None
    return None


def py_timestamp(t) -> datetime:
    return datetime.strptime(t[0:19], '%Y-%m-%dT%H:%M:%S')


def py_date(d) -> date:
    return datetime.strptime(d, '%Y-%m-%d').date()


def py_time(t) -> time:
    return datetime.strptime(t[0:8], "%H:%M:%S").time()


def scrub(s):
    if '\\t' in s or '\0' in s:
        s = s.replace('\\t', ' ')
        s = s.replace('\0', '')
    return s



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Dict

from gurglefish import FileManager
from gurglefish.DriverManager import DbDriverMeta
from gurglefish.context import Context
from gurglefish.objects.files import LocalTableConfig
from gurglefish.objects.sobject import ColumnMap
from gurglefish.sfapi import SObjectFields, SFClient
from gurglefish.tools import make_arg_list

__author__ = 'mark'


class SFSchemaManager:

    def __init__(self, context: Context):
        self.filters = context.filemgr.get_global_filters() + context.filemgr.get_filters()
        self.context = context
        self.log = logging.getLogger("schema")

    @property
    def driver(self) -> DbDriverMeta:
        return self.context.driver

    @property
    def sfclient(self) -> SFClient:
        return self.context.sfclient

    @property
    def filemgr(self) -> FileManager:
        return self.context.filemgr

    @property
    def storagedir(self) -> str:
        return self.context.filemgr.schemadir

    def inspect(self) -> [Dict]:
        solist: [Dict] = self.sfclient.get_sobject_list()
        solist: [Dict] = [sobj for sobj in solist if self.accept_sobject(sobj)]
        for so in solist:
            #
            # enrich the data with the package name
            #
            name = so['name']
            pos = name.find('__')
            if pos != -1 and pos < len(name) - 5:
                so['package'] = name[0:pos]
            else:
                so['package'] = 'unpackaged'
        return solist

    def prepare_configured_sobjects(self):
        table_config: [LocalTableConfig] = self.context.filemgr.get_configured_tables()
        table_list = [table.name for table in table_config if table.enabled]
        return self.prepare_sobjects(table_list)

    def prepare_sobjects(self, names):
        docs = []
        for name in names:
            try:
                doc = self.sfclient.get_sobject_definition(name)
                docs.append(doc)
            except Exception as ex:
                print('Unable to retrieve {}, skipping'.format(name))
                print(ex)
                raise ex
        return self._process_sobjects(docs)

    def accept_sobject(self, sobj: Dict) -> bool:
        """
        determine if the named sobject is suitable for exporting

        :param sobj:Name of sobject/table
        :return: True or False
        """
        name = sobj['name']

        if len(self.filters) > 0 and name not in self.filters:
            return False

        if sobj['name'].endswith('_del__c'):
            return False

        if sobj['customSetting'] is True or sobj['replicateable'] is False or sobj['updateable'] is False:
            return False
        if name.endswith('__Tag') or name.endswith('__History') or name.endswith('__Feed'):
            return False
        if name[0:4] == 'Apex' or name in ('scontrol', 'weblink', 'profile'):
            return False
        return True

    def _process_sobjects(self, solist):
        self.sodict = dict([(so['name'], so) for so in solist])
        sobject_names = set([so['name'].lower() for so in solist])
        for new_sobject_name in sorted(sobject_names):
            self.create_table(new_sobject_name)

    def create_table(self, sobject_name: str):
        new_sobject_name = sobject_name.lower()

        fields: SObjectFields = self.filemgr.get_sobject_fields(sobject_name)
        if fields is None:
            fields: SObjectFields = self.sfclient.get_field_list(new_sobject_name)
            self.filemgr.save_sobject_fields(sobject_name, fields)

        table_name, fieldlist, create_table_dml = self.driver.make_create_table(fields, new_sobject_name)
        select = self.driver.make_select_statement([field.sobject_field for field in fieldlist],
                                                   new_sobject_name)

        parser = self.driver.make_transformer(new_sobject_name, table_name, fieldlist)

        self.filemgr.save_sobject_fields(new_sobject_name, fields)
        self.filemgr.save_sobject_transformer(new_sobject_name, parser)
        self.filemgr.save_sobject_map(new_sobject_name, fieldlist)
        self.filemgr.save_table_create(new_sobject_name, create_table_dml + ';\n\n')
        self.filemgr.save_sobject_query(new_sobject_name, select)

        #
        # now create the table, if needed
        #

        try:
            if not self.driver.table_exists(sobject_name):
                self.log.info(f'  creating {sobject_name}')
                self.driver.exec_ddl(create_table_dml)
                self.log.info(f'  creating indexes')
                self.driver.maintain_indexes(sobject_name, fields)
        except Exception as ex:
            print(ex)
            raise ex

    def update_sobject_definition(self, sobject_name: str, allow_add=True, allow_drop=True):
        sobject_name = sobject_name.lower()

        sobj_columns: SObjectFields = self.sfclient.get_field_list(sobject_name)
        table_columns = self.driver.get_db_columns(sobject_name)

        #
        # check for added/dropped columns
        #
        table_field_names = set([tbl['column_name'] for tbl in table_columns])
        new_field_names = sobj_columns.names() - table_field_names
        dropped_fields = table_field_names - sobj_columns.names()

        if len(new_field_names) > 0:
            if not allow_add:
                self.log.warning(f'  new column found for {sobject_name}, auto-create disabled, skipping')
            else:
                self.log.info(f'  new columns found, updating table and indexes')
                new_field_defs = [sobj_columns.find(f) for f in new_field_names]
                newfields: [ColumnMap] = self.driver.alter_table_add_columns(new_field_defs, sobject_name)
                if len(newfields) > 0:
                    self.driver.maintain_indexes(sobject_name, SObjectFields(new_field_defs))

                    fieldmap: [ColumnMap] = self.filemgr.get_sobject_map(sobject_name)
                    fieldmap.extend(newfields)
                    self.filemgr.save_sobject_map(sobject_name, fieldmap)
                    select = self.driver.make_select_statement([field['sobject_field'] for field in fieldmap],
                                                               sobject_name)
                    self.filemgr.save_sobject_query(sobject_name, select)
                    parser = self.driver.make_transformer(sobject_name, sobject_name, fieldmap)
                    self.filemgr.save_sobject_transformer(sobject_name, parser)

                    self.filemgr.save_sobject_fields(sobject_name, [f for f in sobj_columns.values()])

        if len(dropped_fields) > 0:
            if not allow_drop:
                self.log.warning(f'  dropped column detected for {sobject_name}, auto-drop disabled, skipping')
                # do not allow sync until field(s) allowed to be dropped
                return False
            fieldmap: [ColumnMap] = self.filemgr.get_sobject_map(sobject_name)
            newlist: [ColumnMap] = list()
            for item in fieldmap:
                if item.sobject_field in dropped_fields:
                    pass
                else:
                    newlist.append(item)
            fieldmap = newlist
            self.log.info(f'  dropped column(s) detected')
            self.driver.alter_table_drop_columns(dropped_fields, sobject_name)
            self.filemgr.save_sobject_map(sobject_name, fieldmap)
            select = self.driver.make_select_statement([field.sobject_field for field in fieldmap], sobject_name)
            self.filemgr.save_sobject_query(sobject_name, select)
            parser = self.driver.make_transformer(sobject_name, sobject_name, fieldmap)
            self.filemgr.save_sobject_transformer(sobject_name, parser)

            self.filemgr.save_sobject_fields(sobject_name, sobj_columns)
        return True

    def initialize_config(self, envname: str):
        if self.filemgr.get_configured_tables() is not None:
            self.log.error('Initialization halted, config.json already exists. '
                           'Remove file and tables manually to start over')
            exit(1)
        sobject_list: [Dict] = self.inspect()
        sobjectconfig = []
        for sobject in sobject_list:
            sobjectconfig.append(LocalTableConfig({'name': sobject['name'].lower(), 'enabled': False}))
        self.filemgr.save_configured_tables(sobjectconfig)
        self.log.info(f'Initial configuration created for {envname}')

    def enable_table_sync(self, table_names: [str], flag: bool):
        table_config: [LocalTableConfig] = self.filemgr.get_configured_tables()
        to_enable = [a.lower() for a in make_arg_list(table_names)]
        for entry in table_config:
            if entry.name in to_enable:
                self.log.info(f"Setting {entry.name} sync to {flag}")
                entry.enabled = flag
        self.filemgr.save_configured_tables(table_config)



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict


class LocalTableConfig(object):

    def __init__(self, adict: Dict):
        self.item = adict

    @property
    def dict(self):
        return self.item

    @property
    def name(self):
        return self.item['name']

    @property
    def enabled(self):
        return self.item.get('enabled', False)

    @enabled.setter
    def enabled(self, val):
        self.item['enabled'] = val

    @property
    def auto_drop_columns(self):
        return self.item.get('auto_drop_columns', True)

    @property
    def auto_create_columns(self):
        return self.item.get('auto_create_columns', True)

    @property
    def auto_scrub(self) -> str:
        return self.item.get('auto_scrub', "daily")

    @property
    def sync_schedule(self):
        return self.item.get('sync_schedule', 'auto')

    @property
    def package_name(self):
        return self.item.get('package', None)

    @property
    def use_bulkapi(self) -> bool:
        return self.item.get('bulkapi', False)



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import gzip
import json
import logging
import operator
import os
import string
import sys
from typing import List, Dict

import psycopg2
import psycopg2.extras
from fastcache import lru_cache
from psycopg2._psycopg import connection, cursor

from gurglefish import FileManager
from gurglefish import config
from gurglefish import tools
from gurglefish.DriverManager import DbDriverMeta, GetDbTablesResult, DbNativeExporter
from gurglefish.objects.connections import ConnectionConfig
from gurglefish.context import Context
from gurglefish.objects.sobject import SObjectField, SObjectFields, ColumnMap


class NativeExporter(DbNativeExporter):

    def __init__(self, sobject: str, db: DbDriverMeta, filemgr: FileManager, just_sample=False, timestamp=None):
        self.sobject_name = sobject.lower()
        self.dbdriver = db
        self.query = None
        self.export_file = None
        self.tablefields = None
        self.xlate_handler = filemgr.load_translate_handler(self.sobject_name)
        self.log = logging.getLogger('exporter')

        fieldlist: [ColumnMap] = filemgr.get_sobject_map(self.sobject_name)
        self.fieldmap = dict((f.db_field.lower(), f) for f in fieldlist)

        self.tablefields: Dict = self.dbdriver.get_table_fields(self.sobject_name)
        self.tablefields: List = sorted(self.tablefields.values(), key=operator.itemgetter('ordinal_position'))
        soqlfields = [fm.sobject_field for fm in self.fieldmap.values()]

        self.query = 'select {} from {}'.format(','.join(soqlfields), self.sobject_name)
        if timestamp is not None:
            self.query += ' where SystemModStamp > {0}'.format(tools.sf_timestamp(timestamp))
        if just_sample:
            self.log.info('sampling 500 records max')
            self.query += ' limit 500'
        self.counter = 0
        self.export_file = gzip.open(os.path.join(filemgr.exportdir, self.sobject_name + '.exp.gz'), 'wb', compresslevel=6)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.export_file is not None:
            self.export_file.close()
            self.export_file = None

    def soql(self) -> str:
        return self.query

    def write(self, rec: Dict):
        transformed: Dict = self.xlate_handler.parse(rec)
        record = NativeExporter.format_for_export(transformed, self.tablefields, self.fieldmap)
        self.export_file.write(record)
        self.counter += 1

    @staticmethod
    def format_for_export(trec: Dict, tablefields: [Dict], fieldmap: Dict[str, ColumnMap]):
        parts = []
        for tf in tablefields:
            n = tf['column_name']
            f = fieldmap[n]
            soqlf = f.sobject_field
            if soqlf in trec:
                val = trec[soqlf]
                if val is None:
                    parts.append('\\N')
                else:
                    if isinstance(val, bool):
                        parts.append('True' if val else 'False')
                    elif isinstance(val, datetime.datetime):
                        parts.append(val.isoformat())
                    elif isinstance(val, str):
                        parts.append(NativeExporter._escape(val))
                    else:
                        parts.append(str(val))
            else:
                parts.append('\\N')
        return bytes('\t'.join(parts) + '\n', 'utf-8')

    @staticmethod
    def _escape(val):
        if '\\' in val or '\n' in val or '\r' in val or '\t' in val:
            val = val.replace('\\', '\\\\')
            val = val.replace('\n', '\\n')
            val = val.replace('\r', '\\r')
            val = val.replace('\t', '\\t')
        return val

    def close(self):
        self.export_file.close()
        if sys.stdout.isatty():
            print("\nexported {} records{}".format(self.counter, ' ' * 10))


class Driver(DbDriverMeta):

    def __init__(self):
        self.driver_type = "postgresql"
        self.dbenv = None
        self.db: connection = None
        self.storagedir = None
        self.log = logging.getLogger('dbdriver')
        self.schema_name = None

    def connect(self, dbenv: ConnectionConfig):
        self.dbenv = dbenv
        dbport = dbenv.dbport if dbenv.dbport is not None and len(dbenv.dbport) > 2 else '5432'
        try:
            self.db: connection = psycopg2.connect(
                "dbname='{0}' user='{1}' password='{2}' host='{3}' port='{4}'".format(dbenv.dbname, dbenv.dbuser,
                                                                                      dbenv.dbpass, dbenv.dbhost,
                                                                                      dbport))
            self.storagedir = os.path.join(config.storagedir, 'db', self.dbenv.id)
            self.schema_name = dbenv.schema
            self.verify_db_setup()
            return True
        except Exception as ex:
            self.log.fatal(f'Unable to log into {dbenv.dbname} at {dbenv.dbhost}:{dbport} for user {dbenv.dbuser}')
            self.log.fatal(ex)
            raise ex

    def exec_ddl(self, ddl: str):
        cur = self.db.cursor()
        cur.execute(ddl)
        self.db.commit()
        cur.close()

    @property
    def dbhost(self):
        return self.dbenv.dbhost

    @property
    def dbport(self):
        return self.dbenv.dbport

    @property
    def dbname(self):
        return self.dbenv.dbname

    @property
    def new_map_cursor(self):
        return self.db.cursor(cursor_factory=psycopg2.extras.DictCursor)

    @property
    def cursor(self) -> cursor:
        return self.db.cursor()

    def close(self):
        self.db.close()

    def verify_db_setup(self):
        self.exec_ddl(f'CREATE SCHEMA IF NOT EXISTS {self.schema_name}')
        if not self.table_exists('gf_mdata_sync_stats'):
            ddl = f'create table {self.schema_name}.gf_mdata_sync_stats (' + \
                  '  id         serial primary key, ' + \
                  '  jobid      integer, ' + \
                  '  table_name text not null, ' + \
                  '  inserts    numeric(8) not null, ' + \
                  '  updates    numeric(8) not null, ' + \
                  '  deletes    numeric(8) not null, ' + \
                  '  api_calls  numeric(8) not null, ' + \
                  '  sync_start timestamp not null default now(), ' + \
                  '  sync_end   timestamp not null default now(), ' + \
                  '  sync_since timestamp not null)'
            self.exec_ddl(ddl)
        if not self.table_exists('gf_mdata_schema_chg'):
            ddl = f'create table {self.schema_name}.gf_mdata_schema_chg (' + \
                  '  id         serial primary key, ' + \
                  '  table_name text not null, ' + \
                  '  col_name   text not null, ' + \
                  '  operation  text not null, ' + \
                  '  date_added timestamp not null default now())'
            self.exec_ddl(ddl)
        if not self.table_exists('gf_mdata_sync_jobs'):
            ddl = f'create table {self.schema_name}.gf_mdata_sync_jobs (' + \
                  '  id         serial primary key, ' + \
                  '  date_start timestamp not null default now(),' + \
                  '  date_finish timestamp)'
            self.exec_ddl(ddl)
            self.exec_ddl(f'alter table {self.schema_name}.gf_mdata_sync_stats add constraint ' +
                          'gf_mdata_sync_stats_job_fk foreign key (jobid) references ' +
                          f'{self.schema_name}.gf_mdata_sync_jobs(id) on delete cascade')

    def start_sync_job(self):
        cur = self.cursor
        cur.execute(f'insert into {self.schema_name}.gf_mdata_sync_jobs (date_start) values (%s) returning id',
                    (datetime.datetime.now(),))
        rowid = cur.fetchone()[0]
        cur.close()
        self.db.commit()
        return rowid

    def finish_sync_job(self, jobid):
        cur = self.cursor
        cur.execute(f'update {self.schema_name}.gf_mdata_sync_jobs set date_finish=%s where id=%s',
                    (datetime.datetime.now(), jobid))
        cur.close()

    def insert_sync_stats(self, jobid, table_name, sync_start, sync_end, sync_since, inserts, updates, deletes, api_calls):
        cur = self.cursor
        if sync_since is None:
            sync_since = datetime.datetime(1970, 1, 1, 0, 0, 0)
        dml = f'insert into {self.schema_name}.gf_mdata_sync_stats ' + \
              '(jobid, table_name, inserts, updates, deletes, sync_start, ' + \
              'sync_end, sync_since, api_calls) values (%s,%s,%s,%s,%s,%s,%s,%s,%s)'

        cur.execute(dml, (jobid, table_name, inserts, updates, deletes, sync_start, sync_end, sync_since, api_calls))
        self.db.commit()

    def clean_house(self, date_constraint: datetime):
        cur = self.cursor
        dml = f'delete from {self.schema_name}.gf_mdata_sync_jobs where date_start < %s'
        cur.execute(dml, (date_constraint,))
        self.db.commit()
        cur.close()

    def insert_schema_change(self, table_name: string, col_name: string, operation: string):
        cur = self.cursor
        dml = f'insert into {self.schema_name}.gf_mdata_schema_chg (table_name, col_name, operation) values (%s,%s,%s)'
        cur.execute(dml, [table_name, col_name, operation])
        self.db.commit()
        cur.close()

    def import_native(self, tablename):
        tablename = tablename.lower()

        exportfile = os.path.join(self.storagedir, 'export', tablename + '.exp.gz')
        with self.cursor as cur:
            with gzip.open(exportfile, 'rb') as infile:
                cur.copy_from(infile, tablename)
            self.db.commit()

    def export_native(self, table_name, output_path):
        table_name = table_name.lower()
        with self.cursor as cur:
            with gzip.open(output_path, 'wb', compresslevel=6) as outfile:
                cur.copy_to(outfile, table_name)

    def delete(self, cur, table_name: str, key: str):
        table_name = self.fq_table(table_name)
        try:
            if cur.execute(f"SELECT id from {table_name} where id=%s", [key]) is None:
                return 0
            cur.execute(f'delete from {table_name} where Id=%s', [key])
            return 1
        except Exception as ex:
            self.log.error(f'Deleting record {key} from {table_name}')
            return 0

    def upsert(self, cur, table_name, trec: dict, journal=None):
        assert ('Id' in trec)

        cur.execute("select * from {} where id = '{}'".format(self.fq_table(table_name), trec['Id']))
        tmp_rec = cur.fetchone()
        orig_rec = {}
        index = 0
        if tmp_rec is not None:
            for col in cur.description:
                orig_rec[col[0]] = tmp_rec[index]
                index += 1

        namelist = []
        data = []

        inserted = False
        updated = False

        if len(orig_rec) == 0:
            table_fields = self.get_table_fields(table_name)
            existing_field_names = table_fields.keys()
            for k, v in trec.items():
                k = k.lower()
                if k in existing_field_names:
                    namelist.append(k)
                    data.append(v)

            valueplaceholders = ','.join('%s' for _ in range(len(data)))
            fieldnames = ','.join(namelist)
            sql = 'insert into {0} ({1}) values ({2});'.format(self.fq_table(table_name), fieldnames, valueplaceholders)
            if journal:
                journal.write(bytes('i:{} --> {}\n'.format(sql, json.dumps(data, default=tools.json_serial)), 'utf-8'))
            cur.execute(sql, data)
            inserted = True
        else:
            #
            # use only the changed field values
            #
            pkey = None
            for k, v in trec.items():
                k = k.lower()
                if k == 'id':
                    pkey = v
                    continue
                if k in orig_rec:
                    if orig_rec[k] != v:
                        namelist.append(k)
                        data.append(v)
                #
                # !!!!! FIX DATE/DATETIME PROBLEM
                #

            if len(namelist) == 0:
                #
                # This is a legit case. It is probably due to overlapping lastmodifieddate in sync query where nothing
                # actually changed.
                return False, False

            assert pkey is not None
            sql = 'update {} set '.format(self.fq_table(table_name))
            sets = []
            for name in namelist:
                sets.append(name + r'=%s')
            sql += ','.join(sets)
            sql += " where id = '{}'".format(pkey)
            if journal:
                journal.write(bytes('u:{} --> {}\n'.format(sql, json.dumps(data, default=tools.json_serial)), 'utf-8'))
            cur.execute(sql, data)
            updated = True
        return inserted, updated

    def commit(self):
        self.db.commit()

    def rollback(self):
        self.db.rollback()

    @lru_cache(maxsize=5, typed=False)
    def get_table_fields(self, table_name: str) -> Dict:
        cur = self.new_map_cursor
        sql = "select column_name, data_type, character_maximum_length, ordinal_position " + \
              "from information_schema.columns " + \
              "where table_name = '{}' " + \
              "order by ordinal_position"
        cur.execute(sql.format(table_name))
        columns = cur.fetchall()
        cur.close()
        table_fields = dict()
        for c in columns:
            table_fields[c['column_name']] = {'column_name': c['column_name'], 'data_type': c['data_type'],
                                              'character_maximum_length': c['character_maximum_length'],
                                              'ordinal_position': c['ordinal_position']}
        return table_fields

    def dump_ids(self, table_name: str, output_filename: str):
        cur = self.cursor
        sql = f'select id from {self.schema_name}.{table_name} order by id'
        cur.execute(sql)
        with open(output_filename, 'w') as out:
            for rec in cur:
                out.write(rec[0] + '\n')
        cur.close()

    def record_count(self, table_name: str) -> int:
        table_cursor = self.db.cursor()
        table_cursor.execute('SELECT count(*) FROM {}.{}'.format(self.schema_name, table_name))
        records = table_cursor.fetchone()
        table_cursor.close()
        return records

    def get_db_tables(self) -> List[GetDbTablesResult]:
        table_cursor = self.db.cursor()
        table_cursor.execute(
            "SELECT table_name FROM information_schema.tables WHERE table_schema=%s ORDER BY table_name",
            (self.schema_name,))
        tables = table_cursor.fetchall()
        table_cursor.close()
        result = [GetDbTablesResult(row[0]) for row in tables]
        return result

    def table_exists(self, table_name: str) -> bool:
        table_cursor = self.db.cursor()
        table_cursor.execute(
            "select count(*) from information_schema.tables where table_name = %s and table_schema=%s",
            (table_name, self.schema_name))
        val = table_cursor.fetchone()
        cnt, = val
        return cnt > 0

    def get_db_columns(self, table_name: str) -> List:
        col_cursor = self.new_map_cursor
        col_cursor.execute(
            "select * from information_schema.columns where table_name=%s " +
            "and table_schema=%s " +
            "order by column_name", (table_name, self.schema_name))
        cols = col_cursor.fetchall()
        col_cursor.close()
        return cols

    def _make_column(self, sobject_name: str, field: SObjectField) -> [ColumnMap]:
        """
            returns:
                list(dict(
                    fieldlen, dml, table_name, sobject_name, sobject_field, db_field, fieldtype
                ))
        """
        assert (sobject_name is not None)
        assert (field is not None)
        #        if field is None: return None,None

        sql = ''
        fieldname = field.name
        fieldtype = field.get_type
        fieldlen = field.length
        if fieldtype in ('picklist', 'multipicklist', 'email', 'phone', 'url'):
            sql += 'varchar ({0}) '.format(fieldlen)
        elif fieldtype in ('string', 'encryptedstring', 'textarea', 'combobox'):
            sql += 'text '
        elif fieldtype == 'datetime':
            sql += 'timestamp '
        elif fieldtype == 'date':
            sql += 'date '
        elif fieldtype == 'time':
            sql += 'time '
        elif fieldtype == 'id':
            sql += 'char(15) primary key '
        elif fieldtype == 'reference':
            sql += 'char(15) '
        elif fieldtype == 'boolean':
            sql += 'boolean '
        elif fieldtype == 'double':
            sql += 'numeric ({0},{1}) '.format(field.precision, field.scale)
            fieldlen = field.precision + field.scale + 1
        elif fieldtype == 'currency':
            sql += 'numeric (18,2) '
        elif fieldtype == 'int':
            sql += 'integer '
            fieldlen = 15
        elif fieldtype == 'percent':
            sql += 'numeric '
            fieldlen = 9
        elif fieldtype in ('base64', 'anyType'):  # not implemented yet <<<<<<
            return []
        elif fieldtype == 'address':
            # nothing to handle, this is just an aggregation of fields already exposed for syncing
            return []
        else:
            self.log.error(f'field {fieldname} unknown type {fieldtype} for sobject {sobject_name}')
            raise Exception('field {0} unknown type {1} for sobject {2}'.format(fieldname, fieldtype, sobject_name))

        new_list = [ColumnMap.from_parts(fieldlen, sql, sobject_name, field.name, fieldname, fieldtype)]
        return new_list

    def alter_table_add_columns(self, new_field_defs, sobject_name: str) -> [ColumnMap]:
        ddl_template = "ALTER TABLE {} ADD COLUMN {} {}"
        cur = self.db.cursor()
        newcols = []
        for field in new_field_defs:
            if field.get_type == 'address':
                continue
            col_def: [ColumnMap] = self._make_column(sobject_name, field)
            if len(col_def) == 0:
                continue
            col = col_def[0]
            ddl = ddl_template.format(self.fq_table(sobject_name), col.db_field, col.dml)
            print('    adding column {} to {}'.format(col.db_field, self.fq_table(sobject_name)))
            cur.execute(ddl)
            newcols.append(col)

            # record change to schema
            sql = f'insert into {self.schema_name}.gf_mdata_schema_chg ' +\
                  '(table_name, col_name, operation) values (%s,%s,%s)'
            cur.execute(sql, [sobject_name, col.db_field, 'create'])

        self.db.commit()
        cur.close()
        return newcols

    def alter_table_drop_columns(self, drop_field_names: [str], sobject_name: str):
        ddl_template = 'ALTER TABLE {} DROP COLUMN {}'
        cur = self.db.cursor()
        for field in drop_field_names:
            self.log.info('  dropping column {} from {}'.format(field, sobject_name))
            ddl = ddl_template.format(self.fq_table(sobject_name), field)
            cur.execute(ddl)

            # record change to schema
            sql = 'insert into {}.gf_mdata_schema_chg (table_name, col_name, operation) values (%s,%s,%s)'
            cur.execute(sql.format(self.schema_name), [sobject_name, field, 'drop'])

        self.db.commit()
        cur.close()

    def maintain_indexes(self, sobject_name, field_defs: SObjectFields):
        ddl_template = "CREATE INDEX IF NOT EXISTS {}_{} ON {} ({})"
        cur = self.db.cursor()
        for field in field_defs.values():
            if field.is_externalid or field.is_idlookup or field.name == 'SystemModStamp':
                if field.name != 'id':  # Id is already set as the pkey
                    ddl = ddl_template.format(sobject_name, field.name, self.fq_table(sobject_name), field.name)
                    cur.execute(ddl)
                    self.log.info(f'  created index {sobject_name}_{field.name}')
        self.db.commit()
        cur.close()

    def make_select_statement(self, field_names: [str], sobject_name: str) -> str:
        select = 'select ' + ',\n'.join(field_names) + ' from ' + sobject_name
        return select

    def make_create_table(self, fields: SObjectFields, sobject_name: str) -> (str, [ColumnMap], str):
        sobject_name = sobject_name.lower()
        self.log.info('new sobject: ' + sobject_name)
        tablecols = []
        fieldlist: [ColumnMap] = []

        for field in fields.values():
            m: [ColumnMap] = self._make_column(sobject_name, field)
            if len(m) == 0:
                continue
            for column in m:
                fieldlist.append(column)
                tablecols.append('  ' + column.db_field + ' ' + column.dml)
        sql = ',\n'.join(tablecols)
        return sobject_name, fieldlist, 'create table {0} ( \n{1} )\n'.format(self.fq_table(sobject_name), sql)

    def max_timestamp(self, tablename: str):
        col_cursor = self.db.cursor()
        col_cursor.execute('select max(SystemModStamp) from ' + self.fq_table(tablename))
        stamp, = col_cursor.fetchone()
        col_cursor.close()
        return stamp

    def make_transformer(self, sobject_name, table_name, fieldlist: [ColumnMap]):
        parser = 'from gurglefish.transformutils import id, bl, db, dt, st, ts, tm, inte\n\n'
        parser += 'def parse(rec):\n' + \
                  '  result = dict()\n\n'
        #                  '  def push(name, value):\n' + \
        #                  '    result[name] = value\n\n'

        for field in fieldlist:
            fieldtype = field.field_type
            fieldname = field.sobject_field
            fieldlen = field.fieldlen
            dbfield = field.db_field
            p_parser = ''
            if fieldtype in ('picklist', 'multipicklist', 'string', 'textarea', 'email', 'phone',
                             'url', 'encryptedstring', 'combobox'):
                p_parser = f'result["{dbfield}"] = st(rec, "{fieldname}", fieldlen={fieldlen})\n'
            elif fieldtype == 'datetime':
                p_parser = f'result["{dbfield}"] = ts(rec, "{fieldname}", fieldlen={fieldlen})\n'
            elif fieldtype == 'date':
                p_parser = f'result["{dbfield}"] = dt(rec, "{fieldname}", fieldlen={fieldlen})\n'
            elif fieldtype == 'time':
                p_parser = f'result["{dbfield}"] = tm(rec, "{fieldname}", fieldlen={fieldlen})\n'
            elif fieldtype in ('id', 'reference'):
                p_parser = f'result["{dbfield}"] = id(rec, "{fieldname}", fieldlen={fieldlen})\n'
            elif fieldtype == 'boolean':
                p_parser = f'result["{dbfield}"] = bl(rec, "{fieldname}", fieldlen={fieldlen})\n'
            elif fieldtype in ('double', 'currency', 'percent'):
                p_parser = f'result["{dbfield}"] = db(rec, "{fieldname}", fieldlen={fieldlen})\n'
            elif fieldtype == 'int':
                p_parser = f'result["{dbfield}"] = inte(rec, "{fieldname}", fieldlen={fieldlen})\n'
            elif fieldtype in ('base64', 'anyType'):  # not implemented yet <<<<<<
                return None
            elif fieldtype == 'address':
                # p_parser = 'result["{}"] = stsub(rec, "{}", "{}", fieldlen={})\n'.format(dbfield, fieldname,
                #                                                                         field['subfield'], fieldlen)
                pass

            parser += '  ' + p_parser
        parser += '  return result'
        return parser

    @lru_cache(maxsize=10, typed=False)
    def fq_table(self, tablename):
        return '"{}"."{}"'.format(self.schema_name, tablename)

    @staticmethod
    def _escape(val):
        if '\\' in val or '\n' in val or '\r' in val or '\t' in val:
            val = val.replace('\\', '\\\\')
            val = val.replace('\n', '\\n')
            val = val.replace('\r', '\\r')
            val = val.replace('\t', '\\t')
        return val

    def format_for_export(self, trec: Dict, tablefields: [Dict], fieldmap: Dict[str, ColumnMap]):
        parts = []
        for tf in tablefields:
            n = tf['column_name']
            f = fieldmap[n]
            soqlf = f.sobject_field
            if soqlf in trec:
                val = trec[soqlf]
                if val is None:
                    parts.append('\\N')
                else:
                    if isinstance(val, bool):
                        parts.append('True' if val else 'False')
                    elif isinstance(val, datetime.datetime):
                        parts.append(val.isoformat())
                    elif isinstance(val, str):
                        parts.append(Driver._escape(val))
                    else:
                        parts.append(str(val))
            else:
                parts.append('\\N')
        return bytes('\t'.join(parts) + '\n', 'utf-8')

    def create_exporter(self, sobject_name: str, ctx: Context, just_sample=False, timestamp=None) -> DbNativeExporter:
        exporter = NativeExporter(sobject_name, self, ctx.filemgr, just_sample, timestamp)
        return exporter



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

__author__ = 'Marshall L Smith Jr'

import os
import configparser
from gurglefish.config import storagedir


class ConnectionConfig(object):
    def __init__(self, d):
        self.fields = d

    def to_dict(self):
        return dict([(k, v) for k, v in self.fields.items()])

    @classmethod
    def from_dict(cls, d):
        return ConnectionConfig(d)

    def to_json(self):
        return dict([(k, v) for k, v in self.fields.items()])

    @property
    def id(self):
        return self.fields['id']

    @id.setter
    def id(self, i):
        self.fields['id'] = i

    @property
    def schema(self):
        s = self.fields['schema']
        if s is None or len(s) == 0:
            s = 'public'
        return s

    @schema.setter
    def schema(self, name):
        self.fields['schema'] = name

    @property
    def login(self):
        return self.fields['login']

    @login.setter
    def login(self, l):
        self.fields['login'] = l

    @property
    def password(self):
        return self.fields['password']

    @password.setter
    def password(self, x):
        self.fields['password'] = x

    @property
    def consumer_key(self):
        return self.fields['consumer_key']

    @consumer_key.setter
    def consumer_key(self, value):
        self.fields['consumer_key'] = value

    @property
    def consumer_secret(self):
        return self.fields['consumer_secret']

    @consumer_secret.setter
    def consumer_secret(self, value):
        self.fields['consumer_secret'] = value

    @property
    def authurl(self):
        return self.fields['authurl']

    @authurl.setter
    def authurl(self, value):
        self.fields['authurl'] = value

    @property
    def dbvendor(self):
        return self.fields['dbvendor']

    @dbvendor.setter
    def dbvendor(self, value):
        self.fields['dbvendor'] = value

    @property
    def dbname(self):
        return self.fields['dbname']

    @dbname.setter
    def dbname(self, value):
        self.fields['dbname'] = value

    @property
    def dbuser(self):
        return self.fields['dbuser']

    @dbuser.setter
    def dbuser(self, value):
        self.fields['dbuser'] = value

    @property
    def dbpass(self):
        return self.fields['dbpass']

    @dbpass.setter
    def dbpass(self, value):
        self.fields['dbpass'] = value

    @property
    def dbhost(self):
        return self.fields['dbhost']

    @dbhost.setter
    def dbhost(self, value):
        self.fields['dbhost'] = value

    @property
    def dbport(self):
        return self.fields['dbport']

    @dbport.setter
    def dbport(self, value):
        self.fields['dbport'] = value

    @property
    def threads(self) -> int:
        return min(int(self.fields.get('threads', '1')), 4)


class Connections(object):
    def __init__(self, dbpath=None):
        if dbpath is None:
            dbpath = os.path.join(storagedir, 'connections.ini')
        self.dbpath = dbpath
        if os.path.exists(dbpath):
            config = configparser.ConfigParser()
            config.read(self.dbpath)
            self.data = [ConnectionConfig(config[section]) for section in config.sections()]
        else:
            self.data = []

    def exists(self):
        return os.path.exists(self.dbpath)

    def close(self):
        pass

    def fetch_dblist(self):
        return self.data

    def get_db_env(self, envname) -> Optional[ConnectionConfig]:
        for e in self.data:
            if e.id == envname:
                return e
        return None



#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.
import argparse
import logging.config
import os
import sys
from typing import Dict

from gurglefish import tools
from gurglefish.schema import SFSchemaManager
from gurglefish.sfexport import SFExporter

from gurglefish.sfimport import SFImporter


def main():
    parser = argparse.ArgumentParser(
        epilog='@file arguments designate a file containing actual arguments, one per line')
    group = parser.add_mutually_exclusive_group()
    parser.add_argument("env", help="Environment/DB settings name", metavar="env_name")
    group.add_argument("--sync", help="sync table updates", nargs="*", metavar="sobject|@file")
    group.add_argument("--schema", help="load sobject schema and create tables if missing", nargs="*", metavar="sobject|@file")
    group.add_argument("--export", help="export full sobject data to file", nargs="+", metavar="sobject|@file")
    group.add_argument("--load", help="load/import full table data, table must be empty", nargs="*",
                       metavar="sobject|@file")
    group.add_argument("--dump", help="dump contents of table to file", nargs="+", metavar="table|@file")
    parser.add_argument("--inspect", help="list available sobjects", action="store_true")
    #parser.add_argument("--sample", help="sample data (500 rows)", action="store_true")
    group.add_argument("--init", help="create config.json file for given environment", action="store_true")
    parser.add_argument("--enable", help="enable one or more tables to sync", nargs="+", metavar="sobject|@file")
    parser.add_argument("--disable", help="disable one or more tables from sync", nargs="+", metavar="sobject|@file")
    parser.add_argument("--scrub", help="force scrub of deleted records", action="store_true")
    args = parser.parse_args()

    envname = args.env

    context = tools.setup_env(envname)
    if context is None:
        exit(1)
    logger = logging.getLogger('main')
    schema_mgr = SFSchemaManager(context)

    if args.init:
        schema_mgr.initialize_config(envname)
        sys.exit(0)

    if args.inspect:
        thelist: [Dict] = schema_mgr.inspect()
        for entry in thelist:
            logger.info(entry['name'])
        sys.exit(0)

    if args.enable is not None:
        schema_mgr.enable_table_sync(args.enable, True)
        sys.exit(0)

    if args.disable is not None:
        schema_mgr.enable_table_sync(args.enable, False)
        sys.exit(0)

    if args.sync is not None:
        exp = SFExporter(context)
        exp.sync_tables(schema_mgr, args.scrub)

    if args.schema is not None:
        if len(args.schema) > 0:
            final_args = tools.make_arg_list(args.schema)
            schema_mgr.prepare_sobjects(final_args)
        else:
            schema_mgr.prepare_configured_sobjects()

    if args.export is not None:
        exp = SFExporter(context)
        table_list = tools.make_arg_list(args.export)
        exp.export_tables(table_list, just_sample=args.sample)

    if args.dump is not None:
        table_list = tools.make_arg_list(args.dump)
        for table in table_list:
            export_file = os.path.join(context.filemgr.exportdir, table + '.exp.gz')
            context.dbdriver.export_native(table, export_file)

    if args.load and len(args.load) > 0:
        imp = SFImporter(context, schema_mgr)
        table_list = tools.make_arg_list(args.load)
        for tablename in table_list:
            logger.info('loading {}'.format(tablename))
            count = imp.bulk_load(tablename)
            logger.info('loaded {} records'.format(count))


if __name__ == '__main__':
    main()




#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import datetime
import sys
from multiprocessing import Process, JoinableQueue, Queue, Value

import arrow

from gurglefish import FileManager
from gurglefish import tools
from gurglefish.context import Context
from gurglefish.objects.sobject import ColumnMap
from gurglefish.schema import SFSchemaManager
from gurglefish.objects.files import LocalTableConfig
from gurglefish.sfapi import SFClient, SFQueryTooLarge

__author__ = 'mark'


class ExportThread(Process):
    def __init__(self, queue: Queue, env_name: str):
        super().__init__(daemon=True)
        self.queue = queue
        self.env_name = env_name
        self.ctx = tools.setup_env(env_name)
        self.schema_mgr = SFSchemaManager(self.ctx)
        self.filemgr = self.ctx.filemgr
        self.sfclient = self.ctx.sfclient

    def run(self):
        db = self.ctx.dbdriver
        log = logging.getLogger(self.name)
        try:
            table_configs = self.ctx.filemgr.get_configured_tables()
            while not self.queue.empty():
                job = self.queue.get()
                try:
                    table_name = job['table_name'].lower()
                    just_sample = job['just_sample']

                    this_table: LocalTableConfig = None
                    for table_config in table_configs:
                        if table_config.name == table_name:
                            this_table = table_config
                            break

                    if this_table is None:
                        log.error(f'Configuration for {table_name} not found in config.json - skipping')
                        continue

                    if not db.table_exists(table_name):
                        self.schema_mgr.create_table(table_name)

                    total_size = self.ctx.sfclient.record_count(table_name)

                    if this_table.use_bulkapi:
                        if total_size > 200_000:
                            self.ctx.sfclient.add_header('Sforce-Enable-PKChunking', 'chunkSize=5000')
                        else:
                            self.ctx.sfclient.drop_header('Sforce-Enable-PKChunking')
                        #
                        # Salesforce does an annoying thing - datetime fields retrieve via bulk api are in
                        # millis-since-epoch rather than the usual ISO string format.  So, we need to convert
                        # them back. Get the map and remove everything but datetime fields (for speed).
                        #
                        colmap: [ColumnMap] = self.ctx.filemgr.get_sobject_map(table_name)
                        dtmap: [ColumnMap] = list()
                        for col in iter(colmap):
                            if col.field_type == 'datetime':
                                dtmap.append(col)

                        log.info(f'Exporting {total_size} records in {table_name} using bulk query (may take longer)')
                        with db.create_exporter(table_name, self.ctx, just_sample) as exporter:
                            for rec in self.ctx.sfclient.bulk_query(table_name, exporter.soql()):
                                # replace all the numeric timestamps with correct strings
                                for col in dtmap:
                                    epoch = rec.get(col.sobject_field, None)
                                    if epoch is not None:
                                        dt = datetime.datetime.fromtimestamp(epoch / 1000)
                                        rec[col.sobject_field] = tools.sf_timestamp(dt)
                                exporter.write(rec)
                    else:
                        log.info(f'Exporting {table_name}')
                        with db.create_exporter(table_name, self.ctx, just_sample) as exporter:
                            for rec in self.ctx.sfclient.query(exporter.soql()):
                                exporter.write(rec)

                                if exporter.counter % 5000 == 0 and sys.stdout.isatty():
                                    print('{}: exporting {} records: {:.0f}%\r'.format(exporter.sobject_name,
                                                                               total_size,
                                                                               (exporter.counter / total_size) * 100),
                                                                               end='\r', flush=True)
                finally:
                    self.queue.task_done()
        finally:
            db.close()


class SyncThread(Process):
    def __init__(self, queue: Queue, env_name: str, filemgr: FileManager, sfclient: SFClient, total_calls: Value,
                 scrub=False):
        super().__init__(daemon=True)
        self.queue = queue
        self.filemgr = filemgr
        self.sfclient = sfclient
        self.env_name = env_name

        self.context = tools.setup_env(env_name)
        self.schema_mgr = SFSchemaManager(self.context)
        self.filemgr = self.context.filemgr
        self.sfclient = self.context.sfclient
        self.total_calls: Value = total_calls
        self.force_scrub = scrub

    def run(self):
        db = self.context.dbdriver
        log = logging.getLogger(self.name)
        try:
            while not self.queue.empty():
                try:
                    self.sfclient.calls = 0
                    job = self.queue.get()
                    jobid = job['jobid']
                    tabledef: LocalTableConfig = job['table']
                    sobject_name = tabledef.name.lower()

                    log.info(f'Checking {sobject_name} schema for changes')
                    proceed = self.schema_mgr.update_sobject_definition(sobject_name,
                                                                        allow_add=tabledef.auto_create_columns,
                                                                        allow_drop=tabledef.auto_drop_columns)
                    if not proceed:
                        print(f'sync of {sobject_name} skipped due to warnings')
                        continue

                    timestamp = self.context.dbdriver.max_timestamp(sobject_name)
                    soql = self.context.filemgr.get_sobject_query(sobject_name)

                    xlate_handler = self.filemgr.load_translate_handler(sobject_name)
                    new_sync = False
                    if timestamp is not None:
                        soql += " where SystemModStamp >= {}".format(tools.sf_timestamp(timestamp))
                        soql += " order by SystemModStamp ASC"
                        log.info(f'start sync {sobject_name} changes after {timestamp}')
                    else:
                        soql += ' order by SystemModStamp ASC'
                        log.info(f'start full download of {sobject_name}')
                        new_sync = True
                    with db.cursor as cur:
                        counter = 0
                        # journal = self.filemgr.create_journal(sobject_name)
                        try:
                            sync_start = datetime.datetime.now()
                            inserted = 0
                            updated = 0
                            deleted = 0
                            for rec in self.sfclient.query(soql, not new_sync):
                                del rec['attributes']
                                if rec.get('IsDeleted', False):
                                    deleted += db.delete(cur, sobject_name, rec['Id'][0:15])
                                    continue
                                trec = xlate_handler.parse(rec)

                                try:
                                    i, u = db.upsert(cur, sobject_name, trec, None)
                                    if i:
                                        inserted += 1
                                    if u:
                                        updated += 1
                                except Exception as ex:
                                    # with open('/tmp/debug.json', 'w') as x:
                                    #     x.write(json.dumps(trec, indent=4, default=tools.json_serial))
                                    raise ex

                                if i or u:
                                    counter += 1
                                    if counter % 5000 == 0:
                                        log.info(f'{sobject_name} processed {counter}')
                                    if counter % 10000 == 0:
                                        db.commit()
                            db.commit()

                            # scrub deleted records
                            if tabledef.auto_scrub == "always" or self.force_scrub:
                                deleted += self.scrub_deletes(cur, sobject_name)

                            self.total_calls.value += self.sfclient.calls
                            log.info(f'end sync {sobject_name}: {inserted} inserts, {updated} updates, {deleted} deletes')
                            log.info(f'API calls used for {sobject_name}: {self.sfclient.calls}')

                            if counter > 0:
                                db.insert_sync_stats(jobid, sobject_name, sync_start, datetime.datetime.now(), timestamp,
                                                     inserted, updated, deleted, self.sfclient.calls)
                        except SFQueryTooLarge:
                            log.error(f'Query for {sobject_name} too large for REST API - switch to bulkapi to continue')

                        except Exception as ex:
                            db.rollback()
                            raise ex
                finally:
                    self.queue.task_done()
        finally:
            db.close()

    def scrub_deletes(self, cur, sobject_name: str) -> int:
        db = self.context.dbdriver
        log = logging.getLogger(self.name)
        table_file = os.path.join('/tmp', sobject_name + '.sobject')
        sobject_file = os.path.join('/tmp', sobject_name + '.table')
        try:
            db.dump_ids(sobject_name, table_file)
            self.context.sfapi.dump_ids(sobject_name, sobject_file)

            # lets 'diff' the two lists, finding deleted rows to purge from local database
            with open(table_file, 'r') as local:
                with open(sobject_file, 'r') as remote:
                    left = set(local.readlines())
                    right = set(remote.readlines())
                    ids_to_delete = left - right
                    for i in ids_to_delete:
                        db.delete(cur, sobject_name, i.strip())
            os.unlink(sobject_file)
            os.unlink(table_file)
            return len(ids_to_delete)

        except Exception as ex:
            log.error(ex)
        return 0


class SFExporter:
    def __init__(self, context: Context):
        self.context = context
        self.storagedir = context.filemgr.exportdir
        os.makedirs(self.storagedir, exist_ok=True)
        self.log = logging.getLogger('main')

    def sync_tables(self, schema_mgr: SFSchemaManager, scrub=False):
        table_config: [LocalTableConfig] = self.context.filemgr.get_configured_tables()
        if table_config is None:
            self.log.error('No configuration found - Use --init to create and then edit')
            return
        tablelist: [LocalTableConfig] = [table for table in table_config if table.enabled]
        if len(tablelist) == 0:
            self.log.warning('No tables enabled for sync')
            return
        jobid = self.context.dbdriver.start_sync_job()
        queue: Queue = JoinableQueue()
        try:
            self.log.info('Building table sync queue')
            for table in tablelist:
                tablename = table.name.lower()
                if not self.context.dbdriver.table_exists(tablename):
                    schema_mgr.create_table(tablename)

                queue.put({'jobid': jobid, 'table': table})

            self.log.info(f'Allocating {self.context.env.threads} thread(s)')
            pool: [SyncThread] = list()
            total_api_calls: Value = Value('i', 0)
            for i in range(0, self.context.env.threads):
                job = SyncThread(queue, self.context.envname, self.context.filemgr, self.context.sfclient,
                                 total_api_calls, scrub)
                job.start()
                pool.append(job)

            for t in pool:
                self.log.debug(f'Waiting on thread {t.name}')
                t.join()
            if not queue.empty():
                self.log.warning('All threads finished before queue was drained')
            self.log.info(f"Total API calls used during sync: {total_api_calls.value}")

        finally:
            self.context.dbdriver.finish_sync_job(jobid)
            self.context.dbdriver.clean_house(arrow.now().shift(months=-2).datetime)

    def export_tables(self, table_list: [str], just_sample=False):
        queue: Queue = JoinableQueue()
        for tablename in table_list:
            tablename = tablename.lower()
            queue.put({'table_name': tablename, 'just_sample': just_sample})

        thread_count = min(self.context.env.threads, queue.qsize())
        self.log.info(f'Allocating {thread_count} thread(s)')
        pool: [ExportThread] = list()
        for i in range(0, thread_count):
            job = ExportThread(queue, self.context.envname)
            job.start()
            pool.append(job)
        for t in pool:
            self.log.debug(f'Waiting on thread {t.name}')
            t.join()
        if not queue.empty():
            self.log.warning('All threads finished before queue was drained')



import setuptools
import sys
import re
import os

if sys.version_info < (3, 6):
    print('gurglefish requires at least Python 3.6 to run.')
    sys.exit(1)

with open(os.path.join('gurglefish', '__init__.py'), encoding='utf-8') as f:
    version = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M).group(1)

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='gurglefish',
    version=version,
    python_requires='>=3.6',
    author='Marshall L Smith Jr',
    author_email='marshallsmithjr@gmail.com',
    description='Sync and maintain Salesforce sobjects in a Postgres database',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url='https://github.com/mlsmithjr/gurglefish',
    include_package_data=True,
#    data_files=[('share/doc/gurglefish', ['README.md', 'LICENSE' ]), ('config', ['gurglefish/logging.yml'])],
    packages=setuptools.find_packages(),
    install_requires=['requests==2.21.0', 'psycopg2-binary==2.8', 'fastcache==1.0.2', 'arrow==0.13.1', 'python-dateutil==2.8.0', 'pyyaml==5.1'],
    entry_points={"console_scripts": ["gurglefish=gurglefish.sfarchive:main"]},
    classifiers=[
      'Programming Language :: Python :: 3',
      'Environment :: Console',
      'Topic :: Database',
      'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      'Intended Audience :: System Administrators',
      'Natural Language :: English',
      'Operating System :: POSIX :: Linux',
      'Operating System :: MacOS :: MacOS X',
      'Operating System :: Microsoft :: Windows :: Windows 10',
    ],
    keywords='salesforce sobject database synchronization snapshots postgres postgresql',
)




#    Copyright 2018, 2019 Marshall L Smith Jr
#
#    This file is part of Gurglefish.
#
#    Gurglefish is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gurglefish is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Gurglefish.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Optional, Set


class SFError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class SObjectField(object):
    def __init__(self, field: Dict):
        self.field = field

    @property
    def name(self) -> str:
        return self.field['name']

    @property
    def is_custom(self) -> bool:
        return self.field['custom']

    @property
    def digits(self) -> int:
        return self.field['digits']

    @property
    def label(self) -> str:
        return self.field['label']

    @property
    def length(self) -> int:
        return self.field['length']

    @property
    def precision(self) -> int:
        return self.field['precision']

    @property
    def scale(self) -> int:
        return self.field['scale']

    @property
    def references(self) -> [str]:
        return self.field.get('referenceTo', [])

    @property
    def relationship_name(self) -> str:
        return self.field.get('relationshipName', None)

    @property
    def get_type(self):
        return self.field['type']

    @property
    def is_unique(self) -> bool:
        return self.field['unique']

    @property
    def is_externalid(self) -> bool:
        return self.field['externalId']

    @property
    def is_idlookup(self) -> bool:
        return self.field['idLookup']


class SObjectFields(object):
    def __init__(self, fields):
        self.fields = dict()
        for field in fields:
            if isinstance(field, dict):
                if field['type'] == 'address':
                    continue
                name = field['name']
            else:
                if field.get_type == 'address':
                    continue
                name = field.name
            self.fields[name.lower()] = SObjectField(field)

    def find(self, name: str) -> Optional[SObjectField]:
        return self.fields.get(name.lower(), None)

    def names(self) -> Set:
        return set(self.fields.keys())

    def values(self) -> [SObjectField]:
        return self.fields.values()

    def values_exportable(self) -> [Dict]:
        result = list()
        for f in self.fields.values():
            result.append(f.field)
        return result


class ColumnMap(object):
    def __init__(self, d: Dict):
        self.field = d

    @staticmethod
    def from_parts(fieldlen: int, dml: str, table_name: str, sobject_field: str,
                   db_field: str, fieldtype: str):
        return ColumnMap({'fieldlen': fieldlen, 'dml': dml, 'table_name': table_name,
                          'sobject_field': sobject_field, 'db_field': db_field, 'fieldtype': fieldtype})

    @property
    def fieldlen(self) -> int:
        return self.field['fieldlen']

    @property
    def table_name(self) -> str:
        return self.field['table_name']

    @property
    def sobject_field(self) -> str:
        return self.field['sobject_field']

    @property
    def db_field(self) -> str:
        return self.field['db_field']

    @property
    def field_type(self) -> str:
        return self.field['fieldtype']

    @property
    def dml(self) -> str:
        return self.field['dml']

    def as_dict(self) -> Dict:
        return self.field
